package bufr

import (
	"bufio"
	"io"
	"strings"

	"git.tordarus.net/Tordarus/dstruct"
)

type Reader struct {
	buf *dstruct.Stack[rune]
	src *bufio.Reader
}

func NewReader(r io.Reader) *Reader {
	return &Reader{
		buf: new(dstruct.Stack[rune]),
		src: bufio.NewReader(r),
	}
}

// Rune returns the next rune in r
func (r *Reader) Rune() (rune, error) {
	rn, _, err := r.src.ReadRune()
	if err == nil {
		r.buf.Push(rn)
	}
	return rn, err
}

// UnreadRune unreads the last rune.
// The next read will include the unread rune.
// It returns ErrNothingToUnread if there wasn't any read yet
func (r *Reader) UnreadRune() error {
	if r.buf.Empty() {
		return ErrNothingToUnread.New()
	}

	if err := r.src.UnreadRune(); err == nil {
		r.buf.Pop()
	} else {
		r.src = prependRune(r.buf.Pop(), r.src)
	}

	return nil
}

// UnreadString calls UnreadRune for each rune in str
// The actual runes are irrelevant.
// Only the rune count of str determines the amount of UnreadRune calls.
// The first error occured will be returned immediately.
func (r *Reader) UnreadString(str string) error {
	for range str {
		err := r.UnreadRune()
		if err != nil {
			return err
		}
	}
	return nil
}

// StringWhile reads runes and calls f for each one.
// It returns all runes as a string for which f returned true.
// It stops when f returns false or an error occured.
// The rune for which f returned false will not be unread.
func (r *Reader) StringWhile(f func(rn rune) bool) (string, error) {
	s := new(strings.Builder)

	var err error
	for rn, err := r.Rune(); err == nil && f(rn); rn, err = r.Rune() {
		s.WriteRune(rn)
	}

	return s.String(), err
}

// StringUntil is a shorthand for r.StringWhile(func(rn rune) bool { return !f(rn) })
func (r *Reader) StringUntil(f func(rn rune) bool) (string, error) {
	return r.StringWhile(func(rn rune) bool { return !f(rn) })
}
